# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, re

import pytest
from pytest import xfail

from jenkinsflow.flow import serial, parallel
from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in, build_started_msg


def test_multiple_invocations_serial_same_flow(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('j1', max_fails=0, expect_invocations=2, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='b')
            ctrl1.invoke('j1', password='something else', s1='asdasdasdasdad')


def test_multiple_invocations_serial_same_flow_same_args(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('j1', max_fails=0, expect_invocations=2, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='b')
            ctrl1.invoke('j1', password='a', s1='b')


def test_multiple_invocations_new_flow(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('j1', max_fails=0, expect_invocations=2, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='b')

        with serial(api, timeout=15, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='something else', s1='asdasdasdasdad')


def test_multiple_invocations_new_flow_same_args(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('j1', max_fails=0, expect_invocations=2, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='b')

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='b')


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_multiple_invocations_parallel_same_flow_queued(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        is_hudson = os.environ.get('HUDSON_URL')
        if is_hudson:  # TODO investigate why this test fails in Hudson
            xfail("Doesn't pass when using Hudson")
            return

        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('j1', max_fails=0, expect_invocations=3, expect_order=1, exec_time=3, params=_params)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1', password='a', s1='invocation1')
            ctrl1.invoke('j1', password='b', s1='invocation2')
            ctrl1.invoke('j1', password='b', s1='invocation3')

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Job Invocation-1 (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1",
            "^Job Invocation-2 (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1",
            "^Job Invocation-3 (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1",
            (
                build_started_msg(api, "jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1", 1, invocation_number=1),
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-1 stopped running",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-1 Status IDLE - build: #",
                "^SUCCESS: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-1",

                build_started_msg(api, "jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1", 2, invocation_number=2),
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-2 stopped running",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-2 Status IDLE - build: #",
                "^SUCCESS: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-2",

                build_started_msg(api, "jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1", 3, invocation_number=3),
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-3 stopped running",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-3 Status IDLE - build: #",
                "^SUCCESS: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-3",
            ),
            "^parallel flow: (",
            "job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-1 SUCCESS",
            "job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-2 SUCCESS",
            "job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_queued__j1' Invocation-3 SUCCESS",
            "^)",
        )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_multiple_invocations_parallel_same_flow_no_args_singlequeued(api_type, capsys):
    """
    Jenkins automatically throws away queued builds of parameterless jobs when another build is invoked,
    so that a max of one build can be queued
    """

    with api_select.api(__file__, api_type) as api:
        is_hudson = os.environ.get('HUDSON_URL')
        if is_hudson:  # TODO investigate why this test fails in Hudson
            xfail("Doesn't pass in Hudson")
            return

        api.flow_job()
        num_inv = 20
        api.job('j1', max_fails=0, expect_invocations=num_inv, expect_order=1, exec_time=15)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.1, poll_interval=0.1) as ctrl1:
            for _ in range(0, num_inv):
                ctrl1.invoke('j1')

        sout, _ = capsys.readouterr()

        any_superseeded = re.compile(" +job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-[0-9]+ SUPERSEDED")
        assert lines_in(
            api_type, sout,
            "^Job Invocation-1 (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1",
            "^Job Invocation-5 (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1",
            "^Job Invocation-{} (1/1,1/1): http://x.x/job/jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1".format(num_inv),
            (
                build_started_msg(api, "jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1", 1, invocation_number=1),
                "^SUPERSEDED: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1'",
                "^SUPERSEDED: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1'",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-1 Status RUNNING - build:",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-{} Status QUEUED - ".format(num_inv),

                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-1 stopped running",
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-1 Status IDLE - build: #",
                "^SUCCESS: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-1",

                build_started_msg(api, "jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1", 2, invocation_number=num_inv),
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-{} stopped running".format(num_inv),
                "^job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-{} Status IDLE - build: #".format(num_inv),
                "^SUCCESS: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-{}".format(num_inv),
            ),
            "^parallel flow: (",
            "job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-1 SUCCESS",
            # python3 TODO: *[any_superseeded for _ in range(0, num_inv)]
            any_superseeded,
            any_superseeded,
            any_superseeded,
            "job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' Invocation-{} SUCCESS".format(num_inv),
            "^)",
        )

        # Make sure that first and last are SUCCESS
        re.match("parallel flow: \\(\n *job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' SUCCESS", sout)
        re.match("job: 'jenkinsflow_test__multiple_invocations_parallel_same_flow_no_args_singlequeued__j1' SUCCESS\n *\\)", sout)


def test_multiple_invocations_parallel_same_flow_running(api_type, capsys):
    """Requires job to be setup allowing simultaneous executions"""
    xfail("TODO Note: Test job allowing concurrent builds!")



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function
import os
from os.path import join as jp
from collections import OrderedDict

from jenkinsflow.api_base import BuildResult, Progress, UnknownJobException, ApiInvocationMixin

from .base_test_api import TestJob, TestJenkins
from jenkinsflow.test.cfg import ApiType
from .hyperspeed import HyperSpeed

here = os.path.abspath(os.path.dirname(__file__))


class MockApi(TestJenkins, HyperSpeed):
    job_xml_template = jp(here, 'job.xml.tenjin')
    api_type = ApiType.MOCK

    def __init__(self, job_name_prefix, speedup, public_uri):
        super().__init__(job_name_prefix=job_name_prefix, speedup=speedup)
        self.public_uri = public_uri
        self.username = 'dummy'
        self.password = 'dummy'

    def job(self, name, max_fails, expect_invocations, expect_order, exec_time=None, initial_buildno=None, invocation_delay=0.1, params=None,
            script=None, unknown_result=False, final_result=None, serial=False, print_env=False, flow_created=False, create_job=None,
            disappearing=False, non_existing=False, kill=False, num_builds_to_keep=4, allow_running=False, final_result_use_cli=False,
            set_build_descriptions=()):
        job_name = self.job_name_prefix + name
        assert not self.test_jobs.get(job_name)
        assert isinstance(max_fails, int)
        
        if exec_time is None:
            exec_time = 0.01

        job = MockJob(name=job_name, exec_time=exec_time, max_fails=max_fails, expect_invocations=expect_invocations, expect_order=expect_order,
                      initial_buildno=initial_buildno, invocation_delay=invocation_delay, unknown_result=unknown_result,
                      final_result=final_result, serial=serial, params=params, flow_created=flow_created, create_job=create_job,
                      disappearing=disappearing, non_existing=non_existing, kill=kill, allow_running=allow_running, api=self,
                      final_result_use_cli=final_result_use_cli, set_build_descriptions=set_build_descriptions)
        self.test_jobs[job_name] = job

    def flow_job(self, name=None, params=None):
        # Don't create flow jobs when mocked
        return self.flow_job_name(name)

    # --- Mock API ---

    def poll(self):
        pass

    def quick_poll(self):
        pass

    def queue_poll(self):
        pass

    def get_job(self, name):
        try:
            job = self.test_jobs[name]
            if job.non_existing:
                raise UnknownJobException(name)
            return job
        except KeyError:
            raise UnknownJobException(name)

    def set_build_description(self, description, replace=False, separator='\n', job_name=None, build_number=None):
        pass


class MockJob(TestJob):
    def __init__(self, name, exec_time, max_fails, expect_invocations, expect_order, initial_buildno, invocation_delay, unknown_result,
                 final_result, serial, params, flow_created, create_job, disappearing, non_existing, kill, allow_running, api, final_result_use_cli,
                 set_build_descriptions):
        super().__init__(exec_time=exec_time, max_fails=max_fails, expect_invocations=expect_invocations, expect_order=expect_order,
                                      initial_buildno=initial_buildno, invocation_delay=invocation_delay, unknown_result=unknown_result, final_result=final_result,
                                      serial=serial, print_env=False, flow_created=flow_created, create_job=create_job, disappearing=disappearing,
                                      non_existing=non_existing, kill=kill)
        self.name = name
        self.public_uri = 'http://hupfeldtit.dk/job/' + self.name
        self.build_number = None
        self.initial_build_number = initial_buildno
        self.last_build_number = initial_buildno
        self.params = params
        self._allow_running = allow_running
        self.api = api
        self.final_result_use_cli = final_result_use_cli
        self.set_build_descriptions = set_build_descriptions

        self.just_invoked = False
        self._invocation_url = 0
        self._invocations = OrderedDict()
        self.queued_why = "Why am I queued?"
        self._killed = False
        self._just_killed = False

    def job_status(self):
        latest_build_number = self._get_last_build_number_or_none()
        if self._killed:
            if not self._just_killed:
                return (BuildResult.ABORTED, Progress.IDLE, latest_build_number)
            self._just_killed = False

        if self.start_time <= self.api.time() < self.end_time:
            assert latest_build_number
            return (BuildResult.UNKNOWN, Progress.RUNNING, latest_build_number)

        if self.invocation_time <= self.api.time() < self.start_time:
            return (BuildResult.UNKNOWN, Progress.QUEUED, latest_build_number)

        # Job is finished
        if self.invocation_number <= self.max_fails:
            return (BuildResult.FAILURE, Progress.IDLE, latest_build_number)
        if self.final_result is None:
            return (BuildResult.SUCCESS, Progress.IDLE, latest_build_number)
        return (BuildResult[self.final_result.name], Progress.IDLE, latest_build_number)

    def poll(self):
        # If has been invoked and started running or already (supposed to be) finished
        if self.just_invoked and self.api.time() >= self.start_time:
            self.just_invoked = False
            self.build_number = self.last_build_number + 1 if self.last_build_number else 1
            self._invocations[self._invocation_url - 1].build_number = self.build_number

    def _get_last_build_number_or_none(self):
        self.poll()
        return self.build_number or self.last_build_number

    def _is_running(self):
        return self.start_time <= self.api.time() < self.end_time

    def invoke(self, securitytoken, build_params, cause, description):
        if not self._allow_running:
            if self._is_running():
                print("start_time:", self.start_time, "self.api.time:", self.api.time(), "end_time:", self.end_time)
                assert False

        super().invoke(securitytoken, build_params, cause, description)
        self.invocation_time = self.api.time()
        self.start_time = self.invocation_time + self.invocation_delay
        self.end_time = self.start_time + self.exec_time
        self.just_invoked = True
        self.last_build_number = self.build_number or self.last_build_number
        self.build_number = None

        inv = Invocation(self)
        inv.queued_why = "Why am I queued?"
        self._invocations[self._invocation_url] = inv
        self._invocation_url += 1
        self.poll()
        return inv

    def stop_all(self):
        for inv in self._invocations.values():
            inv.stop(False)
        if self._is_running():
            if not self._killed:
                self._just_killed = True
            self._killed = True

    def update_config(self, config_xml):
        pass

    def __repr__(self):
        return self.name + ", " + super().__repr__()


class Invocation(ApiInvocationMixin):
    def __init__(self, job):
        # _debug("queued_item_path:", queued_item_path)
        self.job = job
        self.build_number = None
        self.queued_why = None
        self._killed = False

    def status(self):
        """Result and Progress info for the invocation"""
        if self._killed:
            return (BuildResult.ABORTED, Progress.IDLE)

        if self.build_number is None:
            return (BuildResult.UNKNOWN, Progress.QUEUED)

        result, progress, _build_number = self.job.job_status()
        return result, progress

    def stop(self, dequeue):
        _, progress, _ = self.job.job_status()
        if progress == Progress.RUNNING and not dequeue:
            self._killed = True



# -*- coding: utf-8 -*-

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os

from pytest import xfail  # pylint: disable=no-name-in-module

from jenkinsflow.flow import serial
from jenkinsflow.rest_api_wrapper import RestkitRestApi

from .framework import api_select


_here = os.path.dirname(os.path.abspath(__file__))


def test_unicode_params_call(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        _params = (('password', '', 'Description'),)
        api.job('job-1', max_fails=0, expect_invocations=1, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('job-1', password='æøå˝$¼@£³¹¶⅝÷«»°¿¦')


def test_unicode_params_defaults(api_type):
    """Job load with unicode"""
    with api_select.api(__file__, api_type, login=True) as api:
        if hasattr(api, 'rest_api') and isinstance(api.rest_api, RestkitRestApi):
            xfail("TODO: this will not work with restkit.")

        api.flow_job()
        _params = (('password', 'æøå˝$¼@£³¹¶⅝÷«»°¿¦', 'Description: ÆØÅEˇé'),)
        api.job('job-1', max_fails=0, expect_invocations=1, expect_order=1, params=_params)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('job-1', password='hello')



# Copyright (c) 2012 - 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import time
import datetime


def log(file, *msg):
    print(*msg)
    sys.stdout.flush()
    print(*msg, file=file)
    file.flush()


def logt(file, *msg):
    now = datetime.datetime.isoformat(datetime.datetime.utcnow())
    log(file, now, *msg)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial

from .framework import api_select


prefixed_jobs = """
serial flow: [
   job: 'top_quick1'
   serial flow: [
      job: 'top_x_quick2-1'
   ]

   serial flow: [
      job: 'top_x_quick2-2'
   ]

   serial flow: [
      job: 'top_x_quick2-3'
   ]

   job: 'top_quick3'
   parallel flow: (
      serial flow: [
         job: 'top_y_z_quick4a'
      ]

      serial flow: [
         job: 'quick4b'
      ]

      job: 'top_y_quick5'
   )

]
"""

def test_prefix(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        def job(name):
            api.job(name, max_fails=0, expect_invocations=0, expect_order=None, params=None)
    
        api.flow_job()
        job('quick1')
        index = 0
        for index in 1, 2, 3:
            job('x_quick2-' + str(index))
        job('quick3')
        job('y_z_quick4')
        job('y_quick5')

        with serial(api, timeout=70, report_interval=3, job_name_prefix='top_', just_dump=True) as ctrl1:
            ctrl1.invoke('quick1')
    
            for index in 1, 2, 3:
                with ctrl1.serial(timeout=20, report_interval=3, job_name_prefix='x_') as ctrl2:
                    ctrl2.invoke('quick2-' + str(index))
    
            ctrl1.invoke('quick3')
    
            with ctrl1.parallel(timeout=40, report_interval=3, job_name_prefix='y_') as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3, job_name_prefix='z_') as ctrl3a:
                    ctrl3a.invoke('quick4a')
                # Reset prefix
                with ctrl2.serial(timeout=40, report_interval=3, job_name_prefix=None) as ctrl3b:
                    ctrl3b.invoke('quick4b')

                ctrl2.invoke('quick5')

        sout, _ = capsys.readouterr()
        assert prefixed_jobs.strip() in sout



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, sys, re
from os.path import join as jp

here = os.path.abspath(os.path.dirname(__file__))
sys.path.extend([jp(here, '../../..'), jp(here, '../../demo')])

from jenkinsflow.test import cfg as test_cfg

from jenkinsflow.unbuffered import UnBuffered
sys.stdout = UnBuffered(sys.stdout)

_file_name_subst = re.compile(r'(_jobs|_test)?\.py')


def api(file_name, api_type, login=None, fixed_prefix=None, url_or_dir=None, fake_public_uri=None, invocation_class=None,
        username=None, password=None):
    """Factory to create either Mock or Wrap api"""
    base_name = os.path.basename(file_name).replace('.pyc', '.py')
    job_name_prefix = _file_name_subst.sub('', base_name)
    func_name = None
    func_num_params = 0
    if fixed_prefix:
        job_name_prefix = fixed_prefix
        file_name = base_name
    elif '_test' in file_name:
        func_name = sys._getframe().f_back.f_code.co_name  # pylint: disable=protected-access
        func_num_params = sys._getframe().f_back.f_code.co_argcount  # pylint: disable=protected-access
        file_name = base_name
        func_name = func_name.replace('test_', '')
        assert func_name[0:len(job_name_prefix)] == job_name_prefix, \
            "Naming standard not followed: " + repr('test_' + func_name) + " defined in file: " + repr(base_name) + " should be 'test_" + job_name_prefix + "_<sub test>'"
        job_name_prefix = 'jenkinsflow_test__' + func_name + '__'
    else:
        job_name_prefix = 'jenkinsflow_demo__' + job_name_prefix + '__'
        file_name = base_name.replace('_jobs', '')

    print()
    print("--- Preparing api for ", repr(job_name_prefix), "---")

    if api_type == 0:
        # Invocation from actual Jenkins flow job calls with api_type == 0
        api_type = test_cfg.ApiType.JENKINS
    print('Using:', api_type)

    url_or_dir = url_or_dir or test_cfg.direct_url(api_type)
    reload_jobs = not test_cfg.skip_job_load() and not fixed_prefix
    pre_delete_jobs = not test_cfg.skip_job_delete()

    import demo_security as security
    if login is None:
        login = security.default_use_login

    if password is not None or username is not None:
        assert password is not None and username is not None
        login = True

    if username is None:
        assert password is None
        username = security.username
        password = security.password

    if api_type == test_cfg.ApiType.JENKINS:
        from .api_wrapper import JenkinsTestWrapperApi
        return JenkinsTestWrapperApi(file_name, func_name, func_num_params, job_name_prefix, reload_jobs, pre_delete_jobs,
                                     url_or_dir, fake_public_uri, username, password, security.securitytoken, login=login,
                                     invocation_class=invocation_class)
    if api_type == test_cfg.ApiType.SCRIPT:
        from .api_wrapper import ScriptTestWrapperApi
        return ScriptTestWrapperApi(file_name, func_name, func_num_params, job_name_prefix, reload_jobs, pre_delete_jobs,
                                    url_or_dir, fake_public_uri, username, password, security.securitytoken, login=login,
                                    invocation_class=invocation_class)
    if api_type == test_cfg.ApiType.MOCK:
        from .mock_api import MockApi
        return MockApi(job_name_prefix, test_cfg.speedup(), test_cfg.direct_url(api_type))

    raise Exception("Unhandled api_type:" + repr(api_type))



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from enum import Enum
from .ordered_enum import OrderedEnum


class BuildResult(OrderedEnum):
    # pylint: disable=no-init
    FAILURE = 0
    ABORTED = 1
    UNSTABLE = 2
    SUCCESS = 3
    SUPERSEDED = 4
    DEQUEUED = 5
    UNKNOWN = 6


class Progress(Enum):
    # pylint: disable=no-init
    RUNNING = 1
    QUEUED = 2
    IDLE = 3


class AuthError(Exception):
    pass


class ClientError(AuthError):
    pass


class UnknownJobException(Exception):
    def __init__(self, job_url, api_ex=None):
        super().__init__("Job not found: " + job_url + (", " + repr(api_ex) if api_ex is not None else ""))


class ApiInvocationMixin():
    def console_url(self):
        return (self.job.public_uri + '/' + repr(self.build_number) + '/console') if self.build_number is not None else None



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, time, os
from os.path import join as jp

from objproxies import ObjectWrapper

from jenkinsflow.api_base import UnknownJobException
from jenkinsflow import script_api
from jenkinsflow import jenkins_api
from jenkinsflow.jobload import update_job_from_template
from jenkinsflow.test import cfg as test_cfg
from jenkinsflow.test.cfg import ApiType

import demo_security as security  # pylint: disable=import-error

from .base_test_api import TestJob, TestJenkins, Jobs as TestJobs
from .config import test_tmp_dir, pseudo_install_dir
from .mock_api import MockJob


here = os.path.abspath(os.path.dirname(__file__))


class WrapperJob(TestJob, ObjectWrapper):
    # NOTE: ObjectWrapper class requires all attributes which are NOT proxied to be declared statically and overridden at instance level
    exec_time = None
    max_fails = None
    expect_invocations = None
    expect_order = None
    unknown_result = None
    final_result = None
    final_result_use_cli = None
    serial = None
    disappearing = None
    non_existing = None
    kill = None

    mock_job = None

    invocation_number = None
    invocation_time = None
    invocation_delay = None
    end_time = None
    actual_order = None

    def __init__(self, jenkins_job, mock_job):
        ObjectWrapper.__init__(self, jenkins_job)
        TestJob.__init__(self, exec_time=mock_job.exec_time, max_fails=mock_job.max_fails,
                         expect_invocations=mock_job.expect_invocations, expect_order=mock_job.expect_order,
                         initial_buildno=mock_job.initial_buildno, invocation_delay=mock_job.invocation_delay,
                         unknown_result=mock_job.unknown_result, final_result=mock_job.final_result, serial=mock_job.serial,
                         print_env=False, flow_created=mock_job.flow_created, create_job=mock_job.create_job, disappearing=mock_job.disappearing,
                         non_existing=mock_job.non_existing, kill=mock_job.kill)

    def invoke(self, securitytoken, build_params, cause, description):
        self.invocation_time = time.time()
        if self.disappearing:
            # Delete the job to fake a job that disappears while a flow is running
            print("Deleting disappearing job:", self.name)
            self.jenkins.delete_job(self.name)
            self.jenkins.poll()

        if self.has_force_result_param:
            build_params = build_params or {}
            if self.invocation_number < self.max_fails:
                build_params['force_result'] = 'FAILURE'
            if self.invocation_number >= self.max_fails:
                build_params['force_result'] = 'SUCCESS' if self.final_result is None else self.final_result.name

        invocation = self.__subject__.invoke(securitytoken, build_params=build_params, cause=cause, description=description)  # pylint: disable=no-member
        TestJob.invoke(self, securitytoken, build_params, cause, description)
        return invocation


class Jobs(TestJobs):
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None
        super().__exit__(exc_type, exc_value, traceback)
        test_jobs = self.api.test_jobs
        for job_name, job in test_jobs.items():
            if not (job.flow_created or job.non_existing):
                self.api._jenkins_job(job_name, job.exec_time, job.params, None, job.print_env, job.create_job,
                                      always_load=job.disappearing, num_builds_to_keep=4, final_result_use_cli=False,
                                      set_build_descriptions=job.set_build_descriptions)


class _TestWrapperApi():
    def __init__(self, file_name, func_name, func_num_params, reload_jobs, pre_delete_jobs, securitytoken, direct_url, fake_public_uri):
        self.file_name = file_name
        self.func_name = func_name
        self.func_num_params = func_num_params
        self.reload_jobs = reload_jobs
        self.pre_delete_jobs = pre_delete_jobs
        self.securitytoken = securitytoken
        self.direct_url = direct_url
        self.fake_public_uri = fake_public_uri
        self.using_job_creator = False

    def _jenkins_job(self, name, exec_time, params, script, print_env, create_job, always_load, num_builds_to_keep,
                     final_result_use_cli, set_build_descriptions):
        # Create job in Jenkins
        if self.reload_jobs or always_load:
            context = dict(
                exec_time=exec_time,
                params=params or (),
                script=script,
                pseudo_install_dir=pseudo_install_dir,
                securitytoken=self.securitytoken,
                username=security.username,
                password=security.password,
                direct_url=self.direct_url,
                print_env=print_env,
                create_job=create_job,
                test_file_name=self.file_name,
                test_tmp_dir=test_tmp_dir, api_type=self.api_type,
                num_builds_to_keep=num_builds_to_keep,
                final_result_use_cli=final_result_use_cli,
                set_build_descriptions=set_build_descriptions,
            )
            update_job_from_template(self.job_loader_jenkins, name, self.job_xml_template, pre_delete=self.pre_delete_jobs, context=context)

    def job(self, name, max_fails, expect_invocations, expect_order, exec_time=None, initial_buildno=None, invocation_delay=0.1, params=None,
            script=None, unknown_result=False, final_result=None, serial=False, print_env=False, flow_created=False, create_job=None, disappearing=False,
            non_existing=False, kill=False, num_builds_to_keep=4, allow_running=False, final_result_use_cli=False,
            set_build_descriptions=()):
        job_name = self.job_name_prefix + name
        assert not self.test_jobs.get(job_name)
        assert isinstance(max_fails, int)

        if create_job or flow_created:
            assert self.using_job_creator

        if exec_time is None:
            exec_time = 0.01

        if max_fails > 0 or final_result:
            params = list(params) if params else []
            force_result_desc = 'Caller can force job to success, fail or unstable.\n' \
                                'If "ABORTED" is specified the job should never run to the end, but be aborted before that.\n' \
                                'If this does not happen, the job will fail.'
            params.append(('force_result', ('SUCCESS', 'FAILURE', 'UNSTABLE', 'ABORTED'), force_result_desc))

        if flow_created or non_existing:
            try:
                print("Deleting job:", job_name)
                self.job_loader_jenkins.delete_job(job_name)
            except UnknownJobException:
                pass
        elif not self.using_job_creator and not non_existing:
            # TODO: Remove and convert all to use job_creator?
            self._jenkins_job(job_name, exec_time, params, script, print_env, create_job=create_job, always_load=disappearing,
                              num_builds_to_keep=num_builds_to_keep, final_result_use_cli=final_result_use_cli,
                              set_build_descriptions=set_build_descriptions)

        job = MockJob(name=job_name, exec_time=exec_time, max_fails=max_fails, expect_invocations=expect_invocations, expect_order=expect_order,
                      initial_buildno=initial_buildno, invocation_delay=invocation_delay, unknown_result=unknown_result, final_result=final_result,
                      serial=serial, params=params, flow_created=flow_created, create_job=create_job, disappearing=disappearing,
                      non_existing=non_existing, kill=kill, allow_running=allow_running, api=self, final_result_use_cli=final_result_use_cli,
                      set_build_descriptions=set_build_descriptions)
        self.test_jobs[job_name] = job

    def flow_job(self, name=None, params=None):
        """
        Creates a flow job
        For running demo/test flow script as jenkins job
        Requires jenkinsflow to be copied to 'pseudo_install_dir' and all jobs to be loaded beforehand (e.g. test.py has been run)
        Returns job name
        """
        name = '0flow_' + name if name else '0flow'
        job_name = (self.job_name_prefix or '') + name
        # TODO Handle script api
        if self.api_type == ApiType.SCRIPT:
            return job_name

        #  Note: Use -B to avoid permission problems with .pyc files created from commandline test
        if self.func_name:
            script = "export PYTHONPATH=" + test_tmp_dir + "\n"
            script += test_cfg.skip_job_load_sh_export_str() + "\n"
            # script += "export " + ApiType.JENKINS.env_name() + "=true\n"  # pylint: disable=no-member
            # Supply dummy args for the py.test fixtures
            dummy_args = ','.join(['0' for _ in range(self.func_num_params)])
            script += sys.executable + " -Bc &quot;import sys; from jenkinsflow.test." + self.file_name.replace('.py', '') + " import *; sys.exit(test_" + self.func_name + "(" + dummy_args + "))&quot;"
        else:
            script = sys.executable + " -B " + jp(pseudo_install_dir, 'demo', self.file_name)
        self._jenkins_job(job_name, exec_time=0.5, params=params, script=script, print_env=False, create_job=None, always_load=False,
                          num_builds_to_keep=4, final_result_use_cli=False, set_build_descriptions=())
        return job_name

    def job_creator(self):
        self.using_job_creator = True
        return Jobs(self)

    # --- Wrapped API ---

    def delete_job(self, job_name):
        TestJenkins.delete_job(self, job_name)
        self.job_loader_jenkins.delete_job(job_name)

    def create_job(self, job_name, config_xml):
        TestJenkins.create_job(self, job_name, config_xml)
        self.job_loader_jenkins.create_job(job_name, config_xml)

    def get_job(self, name):
        job = None
        try:
            job = self.test_jobs.get(name)
            jenkins_job = super().get_job(name)
            if job is None or job.non_existing:
                msg = "InternalError in api_wrapper get_job. Job exists in Jenkins"
                if job is None:
                    msg += ", but not in test_jobs: " + repr(name)
                else:
                    msg += ", but test job has property non_existing: " + repr(name) + ", test job:" + repr(job)
                print(msg, file=sys.stderr)
                raise Exception(msg)

            if isinstance(job, MockJob):
                self.test_jobs[name] = job = WrapperJob(jenkins_job, job)
            assert isinstance(job, WrapperJob)
            job.__subject__ = jenkins_job
            return job
        except UnknownJobException as ex:
            if job and not (job.flow_created or job.non_existing):
                print(ex, file=sys.stderr)
                msg = "In api_wrapper get_job. Job exists in test_jobs, but not in in Jenkins: " + repr(name)
                print(msg, file=sys.stderr)
                raise Exception(msg)
            raise

    def poll(self):
        super().poll()
        if self.fake_public_uri:
            self._public_uri = self.fake_public_uri


class JenkinsTestWrapperApi(_TestWrapperApi, jenkins_api.Jenkins, TestJenkins):
    api_type = ApiType.JENKINS
    job_xml_template = jp(here, 'job.xml.tenjin')

    def __init__(self, file_name, func_name, func_num_params, job_name_prefix, reload_jobs, pre_delete_jobs, direct_url, fake_public_uri,
                 username, password, securitytoken, login, invocation_class):
        TestJenkins.__init__(self, job_name_prefix=job_name_prefix)
        if login:
            jenkins_api.Jenkins.__init__(self, direct_uri=direct_url, job_prefix_filter=job_name_prefix, username=username, password=password,
                                         invocation_class=invocation_class)
        else:
            jenkins_api.Jenkins.__init__(self, direct_uri=direct_url, job_prefix_filter=job_name_prefix, invocation_class=invocation_class)
        self.job_loader_jenkins = jenkins_api.Jenkins(direct_uri=direct_url, job_prefix_filter=job_name_prefix, username=username, password=password)
        _TestWrapperApi.__init__(self, file_name, func_name, func_num_params, reload_jobs, pre_delete_jobs, securitytoken, direct_url, fake_public_uri)


class ScriptTestWrapperApi(_TestWrapperApi, script_api.Jenkins, TestJenkins):
    api_type = ApiType.SCRIPT
    job_xml_template = jp(here, 'job_script.py.tenjin')

    def __init__(self, file_name, func_name, func_num_params, job_name_prefix, reload_jobs, pre_delete_jobs, direct_url, fake_public_uri,
                 username, password, securitytoken, login, invocation_class):
        TestJenkins.__init__(self, job_name_prefix=job_name_prefix)
        if login:
            script_api.Jenkins.__init__(self, direct_uri=direct_url, job_prefix_filter=job_name_prefix, username=username, password=password,
                                        invocation_class=invocation_class)
        else:
            script_api.Jenkins.__init__(self, direct_uri=direct_url, job_prefix_filter=job_name_prefix, invocation_class=invocation_class)
        self.job_loader_jenkins = script_api.Jenkins(direct_uri=direct_url, job_prefix_filter=job_name_prefix, username=username, password=password)
        _TestWrapperApi.__init__(self, file_name, func_name, func_num_params, reload_jobs, pre_delete_jobs, securitytoken, direct_url, fake_public_uri)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from .framework import api_select
from jenkinsflow.flow import serial, FailedChildJobException


def test_hide_password_failed_job(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('passwd_args', max_fails=1, expect_invocations=1, expect_order=1,
                params=(('s1', 'no-secret', 'desc'), ('passwd', 'p2', 'desc'), ('PASS', 'p3', 'desc')))

        with raises(FailedChildJobException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3, secret_params='.*PASS.*|.*pass.*') as ctrl:
                # NOTE: In order to ensure that passwords are not displayed in a stacktrace you must never put a literal password
                # In the last line in the with statement, or in any statement that may raise an exception. You shold not really
                # put clear text paswords in you code anyway :)
                p1, p2, p3 = 'SECRET', 'hemmeligt', 'not_security'
                ctrl.invoke('passwd_args', password=p1, s1='no-secret', passwd=p2, PASS=p3)

        sout, _ = capsys.readouterr()
        assert '******' in sout
        assert 'no-secret' in sout
        assert p1 not in sout
        assert p2 not in sout
        assert p3 not in sout

        assert p1 not in str(exinfo.value)
        assert p2 not in str(exinfo.value)
        assert p3 not in str(exinfo.value)



# Copyright (c) 2012 - 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from .utils import base_url_and_api


def set_build_description(description, replace=False, separator='\n', username=None, password=None, job_name=None, build_number=None, direct_url=None):
    """Utility method to set/append build description on a job build.

    If this is used from inside the hudson job you do not have to specify 'job_name', 'build_number' and 'direct_url'.

    Args:
        description (str): The description to set on the build.
        replace (bool): Replace existing description, if any, instead of appending.
        separator (str): A separator to insert between any existing description and the new 'description' if 'replace' is not specified.
        username (str): User Name for Jenkin authentication with secured Jenkins.
        password (str): Password of Jenkins User.
        job_name (str): Name of the job to modify a build on. Default is os.environ['JOB_NAME'].
        build_number (int): Build Number to modify. . Default is os.environ['BUILD_NUMBER'].
        direct_url (str): Jenkins URL - preferably non-proxied. If not specified, the value of JENKINS_URL or HUDSON_URL environment variables will be used.
    """

    base_url, api = base_url_and_api(direct_url)
    jenkins = api.Jenkins(direct_uri=base_url, username=username, password=password)
    jenkins.set_build_description(description, replace, separator, job_name, build_number)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial
from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in, build_started_msg, build_queued_msg


def test_reporting_queued(api_type, capsys):
    # TODO
    skip_apis = (ApiType.SCRIPT, ApiType.MOCK)

    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        exp_invocations = 2 if api.api_type not in skip_apis else 1
        unknown_result = False if api.api_type not in skip_apis else True
        api.job('j1', max_fails=0, expect_invocations=exp_invocations, expect_order=None, exec_time=10, unknown_result=unknown_result)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke_unchecked('j1')

        print(api.api_type, type(api.api_type))
        if api.api_type in skip_apis:
            return

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, require_idle=False) as ctrl1:
            ctrl1.invoke('j1')

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_queued__j1",
            build_queued_msg(api, "jenkinsflow_test__reporting_queued__j1", 1),
            build_started_msg(api, "jenkinsflow_test__reporting_queued__j1", 2),
        )
